@@ -11,10 +11,16 @@
 
 namespace ixion {
 
-const char* get_formula_function_name(formula_function_t func)
+std::string_view get_formula_function_name(formula_function_t func)
 {
     return formula_functions::get_function_name(func);
 }
 
+formula_function_t get_formula_function_opcode(std::string_view s)
+{
+    return formula_functions::get_function_opcode(s);
 }
+
+}
+
 /* vim:set shiftwidth=4 softtabstop=4 expandtab: */
