@@ -6,6 +6,7 @@
  */
 
 #include "model_parser.hpp"
+#include "app_common.hpp"
 
 #include "ixion/formula.hpp"
 #include "ixion/formula_name_resolver.hpp"
@@ -13,6 +14,9 @@
 #include "ixion/macros.hpp"
 #include "ixion/address_iterator.hpp"
 #include "ixion/dirty_cell_tracker.hpp"
+#include "ixion/cell_access.hpp"
+#include "ixion/config.hpp"
+#include "ixion/cell.hpp"
 
 #include <sstream>
 #include <iostream>
@@ -47,11 +51,6 @@ long to_long(const mem_str_buf& value)
     return ret;
 }
 
-bool to_bool(const mem_str_buf& value)
-{
-    return value == "true";
-}
-
 bool is_separator(char c)
 {
     switch (c)
@@ -85,7 +84,7 @@ mem_str_buf parse_command_to_buffer(const char*& p, const char* p_end)
     return buf;
 }
 
-class string_printer : public std::unary_function<string_id_t, void>
+class string_printer
 {
     const model_context& m_cxt;
     char m_sep;
@@ -110,7 +109,7 @@ public:
 
 void print_section_title(const char* title)
 {
-    std::cout << get_formula_result_output_separator() << std::endl << title << std::endl;
+    std::cout << detail::get_formula_result_output_separator() << std::endl << title << std::endl;
 }
 
 namespace commands {
@@ -127,6 +126,7 @@ enum class type
     mode_init,
     mode_edit,
     mode_result,
+    mode_result_cache,
     mode_table,
     mode_session,
     mode_named_expression,
@@ -146,6 +146,7 @@ const std::vector<map_type::entry> entries =
     { IXION_ASCII("mode init"),             type::mode_init             },
     { IXION_ASCII("mode named-expression"), type::mode_named_expression },
     { IXION_ASCII("mode result"),           type::mode_result           },
+    { IXION_ASCII("mode result-cache"),     type::mode_result_cache     },
     { IXION_ASCII("mode session"),          type::mode_session          },
     { IXION_ASCII("mode table"),            type::mode_table            },
     { IXION_ASCII("print dependency"),      type::print_dependency      },
@@ -178,18 +179,17 @@ model_parser::check_error::check_error(const string& msg) :
 // ============================================================================
 
 model_parser::model_parser(const string& filepath, size_t thread_count) :
-    m_context(),
+    m_context({1048576, 1024}),
     m_table_handler(),
     m_session_handler_factory(m_context),
     mp_table_entry(nullptr),
     mp_name_resolver(formula_name_resolver::get(formula_name_resolver_t::excel_a1, &m_context)),
     m_filepath(filepath),
+    m_strm(detail::load_file_content(m_filepath)),
     m_thread_count(thread_count),
     mp_head(nullptr),
     mp_end(nullptr),
     mp_char(nullptr),
-    m_row_limit(1048576),
-    m_col_limit(1024),
     m_current_sheet(0),
     m_parse_mode(parse_mode_unknown),
     m_print_separator(false),
@@ -198,8 +198,6 @@ model_parser::model_parser(const string& filepath, size_t thread_count) :
     m_context.set_session_handler_factory(&m_session_handler_factory);
     m_context.set_table_handler(&m_table_handler);
 
-    global::load_file_content(m_filepath, m_strm);
-
     mp_head = m_strm.data();
     mp_end = mp_head + m_strm.size();
 }
@@ -226,7 +224,7 @@ void model_parser::parse()
         if (m_print_separator)
         {
             m_print_separator = false;
-            cout << get_formula_result_output_separator() << endl;
+            cout << detail::get_formula_result_output_separator() << endl;
         }
 
         switch (m_parse_mode)
@@ -240,6 +238,9 @@ void model_parser::parse()
             case parse_mode_result:
                 parse_result();
                 break;
+            case parse_mode_result_cache:
+                parse_result_cache();
+                break;
             case parse_mode_table:
                 parse_table();
                 break;
@@ -258,7 +259,7 @@ void model_parser::parse()
 void model_parser::init_model()
 {
     if (m_context.empty())
-        m_context.append_sheet(IXION_ASCII("sheet"), m_row_limit, m_col_limit);
+        m_context.append_sheet("sheet");
 }
 
 void model_parser::parse_command()
@@ -342,6 +343,14 @@ void model_parser::parse_command()
             m_parse_mode = parse_mode_result;
             break;
         }
+        case commands::type::mode_result_cache:
+        {
+            print_section_title("caching formula results");
+
+            m_parse_mode = parse_mode_result_cache;
+            m_print_separator = true;
+            break;
+        }
         case commands::type::mode_edit:
         {
             print_section_title("editing");
@@ -370,7 +379,7 @@ void model_parser::parse_command()
         {
             m_print_separator = true;
             m_parse_mode = parse_mode_named_expression;
-            mp_named_expression = ixion::make_unique<named_expression_type>();
+            mp_named_expression = std::make_unique<named_expression_type>();
             break;
         }
         case commands::type::print_dependency:
@@ -413,17 +422,25 @@ void model_parser::parse_session()
     }
 
     if (cmd == "row-limit")
-        m_row_limit = to_long(value);
+    {
+        rc_size_t ss = m_context.get_sheet_size();
+        ss.row = to_long(value);
+        m_context.set_sheet_size(ss);
+    }
     else if (cmd == "column-limit")
-        m_col_limit = to_long(value);
+    {
+        rc_size_t ss = m_context.get_sheet_size();
+        ss.column = to_long(value);
+        m_context.set_sheet_size(ss);
+    }
     else if (cmd == "insert-sheet")
     {
-        m_context.append_sheet(value.get(), value.size(), m_row_limit, m_col_limit);
-        cout << "sheet: (name: " << value << ", rows: " << m_row_limit << ", columns: " << m_col_limit << ")" << endl;
+        m_context.append_sheet({value.get(), value.size()});
+        cout << "sheet: (name: " << value << ")" << endl;
     }
     else if (cmd == "current-sheet")
     {
-        m_current_sheet = m_context.get_sheet_index(value.get(), value.size());
+        m_current_sheet = m_context.get_sheet_index({value.get(), value.size()});
 
         if (m_current_sheet == invalid_sheet)
         {
@@ -437,7 +454,7 @@ void model_parser::parse_session()
     else if (cmd == "display-sheet-name")
     {
         cout << "display sheet name: " << value << endl;
-        m_print_sheet_name = to_bool(value);
+        m_print_sheet_name = to_bool({value.get(), value.size()});
         m_session_handler_factory.show_sheet_name(m_print_sheet_name);
     }
 }
@@ -457,7 +474,7 @@ void model_parser::parse_init()
 
         formula_tokens_t tokens =
             parse_formula_string(
-                m_context, pos, *mp_name_resolver, cell_def.value.get(), cell_def.value.size());
+                m_context, pos, *mp_name_resolver, {cell_def.value.get(), cell_def.value.size()});
 
         m_context.set_grouped_formula_cells(cell_def.pos, std::move(tokens));
         m_dirty_formula_cells.insert(cell_def.pos);
@@ -478,9 +495,11 @@ void model_parser::parse_init()
             {
                 formula_tokens_t tokens =
                     parse_formula_string(
-                        m_context, pos, *mp_name_resolver, cell_def.value.get(), cell_def.value.size());
+                        m_context, pos, *mp_name_resolver, {cell_def.value.get(), cell_def.value.size()});
 
-                m_context.set_formula_cell(pos, std::move(tokens));
+                auto ts = formula_tokens_store::create();
+                ts->get() = std::move(tokens);
+                m_context.set_formula_cell(pos, ts);
                 m_dirty_formula_cells.insert(pos);
 
                 cout << get_display_cell_string(pos) << ": (f) " << cell_def.value.str() << endl;
@@ -488,14 +507,14 @@ void model_parser::parse_init()
             }
             case ct_string:
             {
-                m_context.set_string_cell(pos, cell_def.value.get(), cell_def.value.size());
+                m_context.set_string_cell(pos, { cell_def.value.get(), cell_def.value.size() });
 
                 cout << get_display_cell_string(pos) << ": (s) " << cell_def.value.str() << endl;
                 break;
             }
             case ct_value:
             {
-                double v = global::to_double(cell_def.value.get(), cell_def.value.size());
+                double v = to_double({cell_def.value.get(), cell_def.value.size()});
                 m_context.set_numeric_cell(pos, v);
 
                 cout << get_display_cell_string(pos) << ": (n) " << v << endl;
@@ -503,7 +522,7 @@ void model_parser::parse_init()
             }
             case ct_boolean:
             {
-                bool b = global::to_bool(cell_def.value.get(), cell_def.value.size());
+                bool b = to_bool({cell_def.value.get(), cell_def.value.size()});
                 m_context.set_boolean_cell(pos, b);
 
                 cout << get_display_cell_string(pos) << ": (b) " << (b ? "true" : "false") << endl;
@@ -531,7 +550,7 @@ void model_parser::parse_edit()
 
         formula_tokens_t tokens =
             parse_formula_string(
-                m_context, pos, *mp_name_resolver, cell_def.value.get(), cell_def.value.size());
+                m_context, pos, *mp_name_resolver, {cell_def.value.get(), cell_def.value.size()});
 
         m_context.set_grouped_formula_cells(cell_def.pos, std::move(tokens));
         m_dirty_formula_cells.insert(cell_def.pos);
@@ -550,7 +569,7 @@ void model_parser::parse_edit()
         {
             // A valid name is given but with empty definition.  Just remove the
             // existing cell.
-            m_context.erase_cell(pos);
+            m_context.empty_cell(pos);
             continue;
         }
 
@@ -560,9 +579,11 @@ void model_parser::parse_edit()
             {
                 formula_tokens_t tokens =
                     parse_formula_string(
-                        m_context, pos, *mp_name_resolver, cell_def.value.get(), cell_def.value.size());
+                        m_context, pos, *mp_name_resolver, {cell_def.value.get(), cell_def.value.size()});
 
-                m_context.set_formula_cell(pos, std::move(tokens));
+                auto ts = formula_tokens_store::create();
+                ts->get() = std::move(tokens);
+                m_context.set_formula_cell(pos, ts);
                 m_dirty_formula_cells.insert(pos);
                 register_formula_cell(m_context, pos);
                 cout << get_display_cell_string(pos) << ": (f) " << cell_def.value.str() << endl;
@@ -570,13 +591,13 @@ void model_parser::parse_edit()
             break;
             case ct_string:
             {
-                m_context.set_string_cell(pos, cell_def.value.get(), cell_def.value.size());
+                m_context.set_string_cell(pos, { cell_def.value.get(), cell_def.value.size() });
                 cout << get_display_cell_string(pos) << ": (s) " << cell_def.value.str() << endl;
             }
             break;
             case ct_value:
             {
-                double v = global::to_double(cell_def.value.get(), cell_def.value.size());
+                double v = to_double({cell_def.value.get(), cell_def.value.size()});
                 m_context.set_numeric_cell(pos, v);
                 cout << get_display_cell_string(pos) << ": (n) " << v << endl;
             }
@@ -594,7 +615,7 @@ void model_parser::parse_result()
     string name_s = res.first.str();
 
     formula_result fres;
-    fres.parse(m_context, res.second.get(), res.second.size());
+    fres.parse({res.second.get(), res.second.size()});
     model_parser::results_type::iterator itr = m_formula_results.find(name_s);
     if (itr == m_formula_results.end())
     {
@@ -608,6 +629,46 @@ void model_parser::parse_result()
         itr->second = fres;
 }
 
+void model_parser::parse_result_cache()
+{
+    parsed_assignment_type res = parse_assignment();
+
+    string name_s = res.first.str();
+
+    formula_result fres;
+    fres.parse({res.second.get(), res.second.size()});
+
+    formula_name_t fnt = mp_name_resolver->resolve(name_s, abs_address_t(m_current_sheet,0,0));
+
+    switch (fnt.type)
+    {
+        case formula_name_t::cell_reference:
+        {
+            abs_address_t pos = std::get<address_t>(fnt.value).to_abs(abs_address_t());
+            formula_cell* fc = m_context.get_formula_cell(pos);
+            if (!fc)
+            {
+                std::ostringstream os;
+                os << name_s << " is not a formula cell";
+                throw model_parser::parse_error(name_s);
+            }
+
+            fc->set_result_cache(fres);
+
+            cout << get_display_cell_string(pos) << ": " << fres.str(m_context) << endl;
+            break;
+        }
+        case formula_name_t::range_reference:
+            throw model_parser::parse_error("TODO: we do not support setting result cache to range just yet.");
+        default:
+        {
+            std::ostringstream os;
+            os << "invalid cell name: " << name_s;
+            throw model_parser::parse_error(os.str());
+        }
+    }
+}
+
 void model_parser::parse_table()
 {
     assert(mp_table_entry);
@@ -620,28 +681,28 @@ void model_parser::parse_table()
     table_handler::entry& entry = *mp_table_entry;
 
     if (name == "name")
-        entry.name = m_context.add_string(value.get(), value.size());
+        entry.name = m_context.add_string({value.get(), value.size()});
     else if (name == "range")
     {
         if (!mp_name_resolver)
             return;
 
         abs_address_t pos(m_current_sheet,0,0);
-        formula_name_t ret = mp_name_resolver->resolve(value.get(), value.size(), pos);
+        formula_name_t ret = mp_name_resolver->resolve({value.get(), value.size()}, pos);
         if (ret.type != formula_name_t::range_reference)
             throw parse_error("range of a table is expected to be given as a range reference.");
 
-        entry.range = to_range(ret.range).to_abs(pos);
+        entry.range = std::get<range_t>(ret.value).to_abs(pos);
     }
     else if (name == "columns")
         parse_table_columns(value);
     else if (name == "totals-row-count")
-        entry.totals_row_count = global::to_double(value.get(), value.size());
+        entry.totals_row_count = to_double({value.get(), value.size()});
 }
 
 void model_parser::push_table()
 {
-    cout << get_formula_result_output_separator() << endl;
+    cout << detail::get_formula_result_output_separator() << endl;
 
     if (!mp_table_entry)
         return;
@@ -679,7 +740,7 @@ void model_parser::parse_named_expression()
 
         formula_name_t name =
             mp_name_resolver->resolve(
-                s.get(), s.size(), abs_address_t(m_current_sheet,0,0));
+                {s.get(), s.size()}, abs_address_t(m_current_sheet,0,0));
 
         if (name.type != formula_name_t::name_type::cell_reference)
         {
@@ -688,13 +749,13 @@ void model_parser::parse_named_expression()
             throw parse_error(os.str());
         }
 
-        mp_named_expression->origin = to_address(name.address).to_abs(abs_address_t(m_current_sheet,0,0));
+        mp_named_expression->origin = std::get<address_t>(name.value).to_abs(abs_address_t(m_current_sheet,0,0));
     }
     else if (res.first == "scope")
     {
         // Resolve it as a sheet name and store the sheet index if found.
         const mem_str_buf& s = res.second;
-        mp_named_expression->scope = m_context.get_sheet_index(s.get(), s.size());
+        mp_named_expression->scope = m_context.get_sheet_index({s.get(), s.size()});
         if (mp_named_expression->scope == invalid_sheet)
         {
             ostringstream os;
@@ -714,15 +775,12 @@ void model_parser::push_named_expression()
 {
     assert(mp_named_expression);
 
-    std::unique_ptr<formula_tokens_t> tokens =
-        ixion::make_unique<formula_tokens_t>(
-            parse_formula_string(
-                m_context, mp_named_expression->origin, *mp_name_resolver,
-                mp_named_expression->expression.data(),
-                mp_named_expression->expression.size()));
+    formula_tokens_t tokens = parse_formula_string(
+        m_context, mp_named_expression->origin, *mp_name_resolver,
+        mp_named_expression->expression);
 
     std::string exp_s = print_formula_tokens(
-        m_context, mp_named_expression->origin, *mp_name_resolver, *tokens);
+        m_context, mp_named_expression->origin, *mp_name_resolver, tokens);
 
     cout << "name: " << mp_named_expression->name << endl;
     cout << "expression: " << exp_s << endl;
@@ -751,14 +809,12 @@ void model_parser::push_named_expression()
 
     if (mp_named_expression->scope == global_scope)
     {
-        m_context.set_named_expression(
-            mp_named_expression->name.data(), mp_named_expression->name.size(), std::move(tokens));
+        m_context.set_named_expression(mp_named_expression->name, std::move(tokens));
     }
     else
     {
         m_context.set_named_expression(
-            mp_named_expression->scope,
-            mp_named_expression->name.data(), mp_named_expression->name.size(), std::move(tokens));
+            mp_named_expression->scope, mp_named_expression->name, std::move(tokens));
     }
 
     mp_named_expression.reset();
@@ -766,7 +822,7 @@ void model_parser::push_named_expression()
 
 void model_parser::print_dependency()
 {
-    std::cout << get_formula_result_output_separator() << std::endl;
+    std::cout << detail::get_formula_result_output_separator() << std::endl;
     std::cout << m_context.get_cell_tracker().to_string() << std::endl;
 }
 
@@ -785,7 +841,7 @@ void model_parser::parse_table_columns(const mem_str_buf& str)
             // Flush the current column name buffer.
             string_id_t col_name = empty_string_id;
             if (!buf.empty())
-                col_name = m_context.add_string(buf.get(), buf.size());
+                col_name = m_context.add_string({buf.get(), buf.size()});
 
             entry.columns.push_back(col_name);
             buf.clear();
@@ -801,7 +857,7 @@ void model_parser::parse_table_columns(const mem_str_buf& str)
 
     string_id_t col_name = empty_string_id;
     if (!buf.empty())
-        col_name = m_context.add_string(buf.get(), buf.size());
+        col_name = m_context.add_string({buf.get(), buf.size()});
 
     entry.columns.push_back(col_name);
 }
@@ -997,10 +1053,6 @@ model_parser::cell_def_type model_parser::parse_cell_definition()
         ret.matrix_value = true;
     }
 
-#if DEBUG_MODEL_PARSER
-    __IXION_DEBUG_OUT__ << "(name='" << ret.name.str() << "'; value='" << ret.value.str() << "')" << endl;
-#endif
-
     if (ret.name.empty())
     {
         if (ret.value.empty())
@@ -1014,19 +1066,19 @@ model_parser::cell_def_type model_parser::parse_cell_definition()
     }
 
     formula_name_t fnt = mp_name_resolver->resolve(
-        ret.name.get(), ret.name.size(), abs_address_t(m_current_sheet,0,0));
+        {ret.name.get(), ret.name.size()}, abs_address_t(m_current_sheet,0,0));
 
     switch (fnt.type)
     {
         case formula_name_t::cell_reference:
         {
-            ret.pos.first = to_address(fnt.address).to_abs(abs_address_t(0,0,0));
+            ret.pos.first = std::get<address_t>(fnt.value).to_abs(abs_address_t(0,0,0));
             ret.pos.last = ret.pos.first;
             break;
         }
         case formula_name_t::range_reference:
         {
-            ret.pos = to_range(fnt.range).to_abs(abs_address_t(0,0,0));
+            ret.pos = std::get<range_t>(fnt.value).to_abs(abs_address_t(0,0,0));
             break;
         }
         default:
@@ -1042,9 +1094,9 @@ model_parser::cell_def_type model_parser::parse_cell_definition()
 
 void model_parser::check()
 {
-    cout << get_formula_result_output_separator() << endl
+    cout << detail::get_formula_result_output_separator() << endl
          << "checking results" << endl
-         << get_formula_result_output_separator() << endl;
+         << detail::get_formula_result_output_separator() << endl;
 
     results_type::const_iterator itr = m_formula_results.begin(), itr_end = m_formula_results.end();
     for (; itr != itr_end; ++itr)
@@ -1054,9 +1106,9 @@ void model_parser::check()
             throw check_error("empty cell name");
 
         const formula_result& res = itr->second;
-        cout << name << " : " << res.str(m_context) << endl;
+        cout << name << ": " << res.str(m_context) << endl;
 
-        formula_name_t name_type = mp_name_resolver->resolve(&name[0], name.size(), abs_address_t());
+        formula_name_t name_type = mp_name_resolver->resolve(name, abs_address_t());
         if (name_type.type != formula_name_t::cell_reference)
         {
             ostringstream os;
@@ -1064,15 +1116,14 @@ void model_parser::check()
             throw std::runtime_error(os.str());
         }
 
-        const formula_name_t::address_type& _addr = name_type.address;
-        abs_address_t addr(_addr.sheet, _addr.row, _addr.col);
+        abs_address_t addr = std::get<address_t>(name_type.value).to_abs(abs_address_t());
+        cell_access ca = m_context.get_cell_access(addr);
 
-        switch (m_context.get_celltype(addr))
+        switch (ca.get_type())
         {
             case celltype_t::formula:
             {
-                const formula_cell* fcell = m_context.get_formula_cell(addr);
-                formula_result res_cell = fcell->get_result_cache();
+                formula_result res_cell = ca.get_formula_result();
 
                 if (res_cell != res)
                 {
@@ -1084,7 +1135,7 @@ void model_parser::check()
             }
             case celltype_t::numeric:
             {
-                double actual_val = m_context.get_numeric_value(addr);
+                double actual_val = ca.get_numeric_value();
                 if (actual_val != res.get_value())
                 {
                     ostringstream os;
@@ -1095,7 +1146,7 @@ void model_parser::check()
             }
             case celltype_t::boolean:
             {
-                bool actual = m_context.get_boolean_value(addr);
+                bool actual = ca.get_boolean_value();
                 bool expected = res.get_value() ? true : false;
                 if (actual != expected)
                 {
@@ -1107,26 +1158,16 @@ void model_parser::check()
             }
             case celltype_t::string:
             {
-                string_id_t str_id = m_context.get_string_identifier(addr);
+                std::string_view actual = ca.get_string_value();
+                const std::string& s_expected = res.get_string();
 
-                if (str_id != res.get_string())
+                if (actual != s_expected)
                 {
-                    const string* ps = m_context.get_string(str_id);
-                    if (!ps)
-                        throw check_error("failed to retrieve a string value for a string cell.");
-
-                    const string* ps_expected = m_context.get_string(res.get_string());
-                    if (!ps_expected)
-                    {
-                        std::ostringstream os;
-                        os << "no string value for the string ID of " << res.get_string();
-                        throw check_error(os.str());
-                    }
-
-                    ostringstream os;
-                    os << "unexpected string result: (expected: '" << *ps_expected << "'; actual: '" << *ps << "')";
+                    std::ostringstream os;
+                    os << "unexpected string result: (expected: '" << s_expected << "'; actual: '" << actual << "')";
                     throw check_error(os.str());
                 }
+
                 break;
             }
             default:
